/*
 *  wiidev_client.c
 *  WiiDev
 *
 *  Created by Ian Rickard on 2/20/07.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <stdio.h>

#include "wiidev_client.h"

WiiDevRef WiiDevNewFromName(const char *name) {
	struct sockaddr_un dst;
	
	dst.sun_family = AF_UNIX;
	snprintf(dst.sun_path,104,"%s/Library/Wii Remotes/%s", getenv("HOME"), name);

	WiiDevRef out = calloc(sizeof(struct WiiDev), 1);
	
	out->sock = socket(AF_UNIX, SOCK_STREAM, 0);
	if (!out->sock) {
		free(out);
		return NULL;
	}
	
	if (connect(out->sock, (struct sockaddr*)&dst, sizeof(dst))) {
		free(out);
		return NULL;
	}
	
	out->xlrzero.x = out->xlrzero.y = out->xlrzero.z = 128;
	out->xlrone.x = out->xlrone.y = out->xlrone.z = 140;
	out->lastUpdateMask = 0;
	
	out->dataCallback = NULL;
	out->memCallback = NULL;
	out->waitingForXlrCal = 0;
	
	int i;
	for(i=0 ; i<0x20 ; i++)
		out->regs[i] = 0;
	
	out->regs[0x12] = 0x0030;
	out->vibe = 0;
	
	return out;
}

void WiiDevRelease(WiiDevRef ref) {
	close(ref->sock);
	free(ref);
}

// handle any incoming data
int WiiDevUpdate(WiiDevRef ref) {
	unsigned char buf[256];
	int err = recv(ref->sock, buf, 1, MSG_WAITALL);
	if (err != 1) {
		printf("recv %d returned %d\n", 1, err);
		return -1;
	}
	int length = buf[0];
	
	err = recv(ref->sock, buf, length, MSG_WAITALL);
	if (err != length) {
		printf("recv %d returned %d\n", length, err);
		return -1;
	}
	printf("recieved packet of length %d\n", length);
	
	return 0;
}
// positive return: number of updates processed.  negative: error.

void WiiDevSetReport(WiiDevRef ref, int report) {
	ref->regs[0x12] = report | 0x0030;
	WiiDevWriteReg(ref, 0x12);
}

void WiiDevEnableContinuousReporting(WiiDevRef ref, int on) {

}

void WiiDevEnableVibration(WiiDevRef ref, int on) {
	ref->vibe = on?1:0;
	WiiDevWriteReg(ref, 0x13); 
}

void WiiDevEnableIR(WiiDevRef ref, int mode) {

}

void WiiDevSetLEDs(WiiDevRef ref, int leds) {
	ref->regs[0x11] = leds<<4;
	WiiDevWriteReg(ref, 0x11);
}

void WiiDevSetMemReadCallback(WiiDevRef ref, WiiDevMemReadCallback callback) {

}

void WiiDevReadMem(WiiDevRef ref, int address, int length) {

}

void WiiDevWriteMem(WiiDevRef ref, int address, WiiDevData data, int length) {

}

void WiiDevWriteReg(WiiDevRef ref, int reg) {
	WiiDevData cmd = {0x52, 0x00, 0x00, 0x00};
	int length;
	
	switch(reg) {
	case 0x12: length = 2;
	default: length = 1;
	}
	
	cmd[1] = reg;
	if (length == 1) {
		cmd[2] = ref->regs[reg];
	} else if (length == 2) {
		cmd[2] = ref->regs[reg]>>8;
		cmd[3] = ref->regs[reg];
	}
	
	if (ref->vibe) cmd[2] |= 0x01;
	
	WiiDevSend(ref, cmd, length+2);
}

void WiiDevSend(WiiDevRef ref, WiiDevData data, int length) {
	unsigned char len = length;
	if (write(ref->sock, &len, 1) != 1)
		perror("write length failed");
	if (write(ref->sock, data, length) != length)
		perror("write payload failed");
}

void WiiDevSetDataReceivedCallback(WiiDevRef ref, WiiDevDataCallback callback) {

}
