/*
 *  wiidev_client.c
 *  WiiDev
 *
 *  Created by Ian Rickard on 2/20/07.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>

#include "wiidev_client.h"

WiiDevRef WiiDevNewFromName(const char *name) {
	struct sockaddr_un dst;
	
	dst.sun_family = AF_UNIX;
	snprintf(dst.sun_path,104,"%s/Library/Wii Remotes/%s", getenv("HOME"), name);

	WiiDevRef out = calloc(sizeof(struct WiiDev), 1);
	
	out->sock = socket(AF_UNIX, SOCK_STREAM, 0);
	if (!out->sock) {
		free(out);
		return NULL;
	}
	
	if (connect(out->sock, (struct sockaddr*)&dst, sizeof(dst))) {
		free(out);
		return NULL;
	}
	
	out->xlrzero.x = out->xlrzero.y = out->xlrzero.z = 256;
	out->xlrone.x = out->xlrone.y = out->xlrone.z = 256+88*2;
	out->lastUpdateMask = 0;
	
	out->dataCallback = NULL;
	out->memCallback = NULL;
	out->waitingForXlrCal = 0;
	
	int i;
	for(i=0 ; i<0x20 ; i++)
		out->regs[i] = 0;
	
	out->regs[0x12] = 0x0030;
	out->vibe = 0;
	
	WiiDevReadMem(out, 0x0020, 8);
	out->waitingForXlrCal = 1;
	
	return out;
}

void WiiDevRelease(WiiDevRef ref) {
	close(ref->sock);
	free(ref);
}

// handle any incoming data
int WiiDevUpdate(WiiDevRef ref) {
	unsigned char buf[256];
	int i;
	
	fd_set readset;
	struct timeval timeout = {0,0};
	FD_ZERO(&readset);
	FD_SET(ref->sock, &readset); 
	
	int updates = 0;
	
	do {
		int err = select(ref->sock+1, &readset, NULL, NULL, &timeout);
		
		if (err == 0) break;
		
		err = recv(ref->sock, buf, 1, MSG_WAITALL);
		if (err != 1) {
			printf("recv %d returned %d\n", 1, err);
			return -1;
		}
		int length = buf[0];
		
		err = recv(ref->sock, buf, length, MSG_WAITALL);
		if (err != length) {
			printf("recv %d returned %d\n", length, err);
			return -1;
		}
		
		if (NULL != ref->dataCallback)
			ref->dataCallback(buf, length);
		
	//	printf("%02X %02X (%d)\n", buf[0], buf[1], length);
	//	printf("recieved packet of length %d from channel %02X\n", length, buf[0]);
		if (length >= 3 && (buf[0] == 0x41 || buf[0] == 0x42) && buf[1] == 0xA1) {
			if ((buf[2]&0xF0) == 0x30) {
				updates++;
				
				// parse button data:
				if (buf[2] != 0x37) {
					ref->buttons = ((int)buf[3])<<8 | buf[4];
					ref->lastUpdateMask = kWiiDevMaskButtons;
				}
				
				// parse XLR data
				switch(buf[2]) {
				case 0x31: case 0x33: case 0x35: case 0x37:
					ref->xlrraw.x = buf[5]<<1 | ((buf[4]&0x40)?1:0);
					ref->xlrraw.y = buf[6]<<1 | ((buf[3]&0x20)?1:0);
					ref->xlrraw.z = buf[7]<<1 | ((buf[3]&0x40)?1:0);
					ref->xlr.x = (ref->xlrraw.x-ref->xlrzero.x) / (float)(ref->xlrone.x - ref->xlrzero.x);
					ref->xlr.y = (ref->xlrraw.y-ref->xlrzero.y) / (float)(ref->xlrone.y - ref->xlrzero.y);
					ref->xlr.z = (ref->xlrraw.z-ref->xlrzero.z) / (float)(ref->xlrone.z - ref->xlrzero.z);
					ref->lastUpdateMask = kWiiDevMaskXLR;
				break;
					case 0x3e:
				break;
					case 0x3f:
				break;
				}
				
				// parse IR data
				switch(buf[2]) {
				case 0x33:
					for(i=0 ; i<4 ; i++) {
						ref->irdata[i].x = buf[ 8+3*i] + ((buf[10+3*i] & 0x30)<<4);
						ref->irdata[i].y = buf[ 9+3*i] + ((buf[10+3*i] & 0xC0)<<2);
						ref->irdata[i].s = buf[10+3*i] & 0x0F;
						ref->irdata[i].live = ref->irdata[i].s != 0x0F;
					}
				//	for(i=0 ; i<12 ; i+=3)
				//		printf("%02X%02X%02X ", buf[i+8], buf[i+9], buf[i+10]);
					ref->lastUpdateMask = kWiiDevMaskIR;
					break;
				case 0x36: case 0x37:
					break;
				case 0x3e:
					for(i=0 ; i<18 ; i++)
						ref->irRaw[i] = buf[6+i];
					break;
				case 0x3f:
					for(i=0 ; i<18 ; i++)
						ref->irRaw[18+i] = buf[6+i];
				//	for(i=0 ; i<36 ; i+=3)
				//		printf("%02X%02X%02X ", ref->irRaw[i+0], ref->irRaw[i+1], ref->irRaw[i+2]);
				//	printf("\n");
					ref->lastUpdateMask = kWiiDevMaskIRFull;
					break;
				}
			} else if (buf[2] == 0x21) {
				int readlen = (buf[5]>>4)+1;
				int addr = (buf[6]<<8) | buf[7];
				
			//	printf("addr %X len %d\n", addr, readlen);
				
				if (readlen == 8 && addr == 0x20 && ref->waitingForXlrCal) {
					
					ref->xlrzero.x = buf[8]<<1;
					ref->xlrzero.y = buf[9]<<1;
					ref->xlrzero.z = buf[10]<<1;
					ref->xlrone.x = buf[12]<<1;
					ref->xlrone.y = buf[13]<<1;
					ref->xlrone.z = buf[14]<<1;
					
					ref->waitingForXlrCal = 0;
				} else if (NULL != ref->memCallback) {
					ref->memCallback(addr, buf+8, readlen);
				}
				
			} else {
				printf("unknown input report %02X\n", buf[2]);
			}
		}
	} while(0);
	
	return updates;
}
// positive return: number of updates processed.  negative: error.

void WiiDevSetReport(WiiDevRef ref, int report, int continuous) {
	ref->regs[0x12] = report;
	
	if (continuous)
		ref->regs[0x12] |= 0x0400;
	
	WiiDevWriteReg(ref, 0x12);
	
	switch(ref->regs[0x12]&0xFF) {
	case kWiiDevReportBXIe:
		WiiDevSetIRMode(ref, kWiiDevIRModeExtended);
		break;
	case kWiiDevReportBIbE9: case kWiiDevReportBXIbE6:
		WiiDevSetIRMode(ref, kWiiDevIRModeBasic);
		break;
	case kWiiDevReportBXIf:
		WiiDevSetIRMode(ref, kWiiDevIRModeFull);
		break;
	default:
		WiiDevSetIRMode(ref, kWiiDevIRModeOff);
		break;
	}
}

void WiiDevSetIRMode(WiiDevRef ref, int mode) {
	if (kWiiDevIRModeOff != mode) {
		ref->regs[0x13] |= 0x04;
		ref->regs[0x1A] |= 0x04;
	} else {
		ref->regs[0x13] &= ~0x04;
		ref->regs[0x1A] &= ~0x04;
	}
	usleep(10000);
	WiiDevWriteReg(ref, 0x13);
	WiiDevWriteReg(ref, 0x1A);
	
	if (kWiiDevIRModeOff != mode) {
		WiiDevWriteMem(ref, 0x04B00030, (WiiDevData){0x01}, 1);
		usleep(10000);
		WiiDevWriteMem(ref, 0x04B00030, (WiiDevData){0x08}, 1);
		usleep(10000);
		WiiDevWriteMem(ref, 0x04B00000, (WiiDevData){0, 0, 0, 0, 0, 0, 0x90, 0, 0x41}, 9);
		WiiDevWriteMem(ref, 0x04B0001A, (WiiDevData){0x40, 0x00}, 2);
	//	WiiDevWriteMem(ref, 0x04B00000, (WiiDevData){0x02, 0, 0, 0x71, 0, 0, 0xAA, 0, 0x64}, 9);
	//	WiiDevWriteMem(ref, 0x04B0001A, (WiiDevData){0x63, 0x03}, 2);
		//WiiDevWriteMem(ref, 0x04B00000, (WiiDevData){0, 0, 0, 0, 0, 0, 0x90, 0, 0x60}, 9);
		//WiiDevWriteMem(ref, 0x04B0001A, (WiiDevData){0x40, 0x01}, 2);
		usleep(10000);
		WiiDevWriteMem(ref, 0x04B00033, (WiiDevData){mode}, 1);
		usleep(10000);
		WiiDevWriteMem(ref, 0x04B00030, (WiiDevData){0x08}, 1);
		usleep(10000);
	}
}

void WiiDevEnableVibration(WiiDevRef ref, int on) {
	ref->vibe = on?1:0;
	WiiDevWriteReg(ref, 0x13); 
}

void WiiDevSetLEDs(WiiDevRef ref, int leds) {
	ref->regs[0x11] = leds<<4;
	WiiDevWriteReg(ref, 0x11);
}

void WiiDevSetMemReadCallback(WiiDevRef ref, WiiDevMemReadCallback callback) {
	ref->memCallback = callback;
}

void WiiDevReadMem(WiiDevRef ref, int address, int length) {
	WiiDevData cmd = {
		0x52, 0x17,
		0,0,0,0,
		0,0};
	
	cmd[2] = (address>>24)&0x000000FF;
	cmd[3] = (address>>16)&0x000000FF;
	cmd[4] = (address>> 8)&0x000000FF;
	cmd[5] = (address>> 0)&0x000000FF;
	
	cmd[6] = (length>>8)&0x00FF;
	cmd[7] = (length>>0)&0x00FF;
	
	WiiDevSend(ref, cmd, sizeof(cmd));
}

void WiiDevWriteMem(WiiDevRef ref, int address, WiiDevData data, int length) {
	WiiDevData cmd = {0x52,0x16,
		0,0,0,0,
		0,
		0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0};
	
	cmd[2] = (address>>24)&0x000000FF;
	cmd[3] = (address>>16)&0x000000FF;
	cmd[4] = (address>> 8)&0x000000FF;
	cmd[5] = (address>> 0)&0x000000FF;
	
	cmd[6] = length;
	
	memcpy(cmd+7, data, length);
	WiiDevSend(ref, cmd, sizeof(cmd));
}

void WiiDevWriteReg(WiiDevRef ref, int reg) {
	WiiDevData cmd = {0x52, 0x00, 0x00, 0x00};
	int length;
	
	switch(reg) {
	case 0x12:
		length = 2;
		break;
	default:
		length = 1;
		break;
	}
	
	cmd[1] = reg;
	if (length == 1) {
		cmd[2] = ref->regs[reg];
		printf("writing reg %02X = %02X\n", reg, cmd[2]);
	} else if (length == 2) {
		cmd[2] = ref->regs[reg]>>8;
		cmd[3] = ref->regs[reg];
		printf("writing reg %02X = %02X%02X\n", reg, cmd[2], cmd[3]);
	}
	
	if (ref->vibe) cmd[2] |= 0x01;
	
	WiiDevSend(ref, cmd, length+2);
}

void WiiDevSend(WiiDevRef ref, WiiDevData data, int length) {
	unsigned char len = length;
	if (write(ref->sock, &len, 1) != 1)
		perror("write length failed");
	if (write(ref->sock, data, length) != length)
		perror("write payload failed");
}

void WiiDevSetDataReceivedCallback(WiiDevRef ref, WiiDevDataCallback callback) {
	ref->dataCallback = callback;
}
