/*
 *  wiidev_client.h
 *  WiiDev
 *
 *  Created by Ian Rickard on 2/20/07.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#ifndef WIIDEV_H
#define WIIDEV_H

#ifdef __cplusplus
extern "C" {
#endif

typedef unsigned char WiiDevData[];
typedef void (*WiiDevMemReadCallback)(int address, const WiiDevData data, int length);
typedef void (*WiiDevDataCallback)(WiiDevData data, int length);

typedef struct WiiDev {
	struct {
		int x, y, z;
	} xlrraw;
	struct {
		int x, y, z;
	} xlrzero;
	struct {
		int x, y, z;
	} xlrone;
	struct {
		float x, y, z;
	} xlr;
	int buttons;
	struct {
		int live, x, y, s;
	} irdata[4];
	unsigned char irRaw[36];
	
	int lastUpdateMask;
	
	// private:
	int sock;
	WiiDevDataCallback dataCallback;
	WiiDevMemReadCallback memCallback;
	int waitingForXlrCal;
	int regs[0x20];
	int vibe;
} *WiiDevRef;

enum {
	WiiDevButton1 = 0x0001,
	WiiDevButton2 = 0x0002,
	WiiDevButtonB = 0x0004,
	WiiDevButtonA = 0x0008,
	WiiDevButtonMinus = 0x0010,
	WiiDevButtonHome = 0x0080,
	WiiDevButtonL = 0x0100,
	WiiDevButtonR = 0x0200,
	WiiDevButtonD = 0x0400,
	WiiDevButtonU = 0x0800,
	WiiDevButtonPlus = 0x1000,
};

enum {
	kWiiDevMaskButtons = 1<<0,
	kWiiDevMaskXLR = 1<<1,
	kWiiDevMaskIR = 1<<2,
	kWiiDevMaskIRFull = 1<<3
};

enum {
	kWiiDevIRModeOff=0,
	kWiiDevIRModeBasic=1,
	kWiiDevIRModeExtended=3,
	kWiiDevIRModeFull=5
};

enum {
	// B  = Buttons
	// X  = XLR
	// Ib = basic IR
	// Ie = extended IR
	// If = full IR
	// E# = extension (# bytes)
	kWiiDevReportB = 0x30,
	kWiiDevReportBX = 0x31,
	kWiiDevReportBE8 = 0x32,
	kWiiDevReportBXIe = 0x33,
	kWiiDevReportBE19 = 0x34,
	kWiiDevReportBXE16 = 0x35,
	kWiiDevReportBIbE9 = 0x36,
	kWiiDevReportBXIbE6 = 0x37,
	kWiiDevReportBE21 = 0x3d,
	kWiiDevReportBXIf = 0x3e // and 3f
};

WiiDevRef WiiDevNewFromName(const char *name);
void WiiDevRelease(WiiDevRef ref);

// handle any incoming data
int WiiDevUpdate(WiiDevRef ref);
// positive return: number of updates processed.  negative: error.

void WiiDevSetReport(WiiDevRef ref, int report, int continuous);
void WiiDevEnableVibration(WiiDevRef ref, int on);
void WiiDevSetLEDs(WiiDevRef ref, int leds);

void WiiDevSetIRMode(WiiDevRef ref, int mode); // don't call directly, set report type

void WiiDevSetMemReadCallback(WiiDevRef ref, WiiDevMemReadCallback callback);
void WiiDevReadMem(WiiDevRef ref, int address, int length);
void WiiDevWriteMem(WiiDevRef ref, int address, WiiDevData data, int length);


// low level
void WiiDevWriteReg(WiiDevRef ref, int reg);
void WiiDevSend(WiiDevRef ref, WiiDevData data, int length);
void WiiDevSetDataReceivedCallback(WiiDevRef ref, WiiDevDataCallback callback);

#ifdef __cplusplus
}
#endif


#endif